import os.path
import shutil

from click.testing import CliRunner

from tidy3d.web.cli import tidy3d_cli
from tidy3d.web.cli.app import CONFIG_FILE


def test_tidy3d_cli():

    if os.path.exists(CONFIG_FILE):
        shutil.move(CONFIG_FILE, f"{CONFIG_FILE}.bak")

    runner = CliRunner()
    result = runner.invoke(tidy3d_cli, ["configure"], input="apikey")

    assert result.exit_code == 0
    if os.path.exists(CONFIG_FILE):
        os.remove(CONFIG_FILE)

    if os.path.exists(f"{CONFIG_FILE}.bak"):
        shutil.move(f"{CONFIG_FILE}.bak", CONFIG_FILE)



"""Interface to run several jobs in batch using asyncio."""
import asyncio
from typing import Dict

import nest_asyncio

from .container import DEFAULT_DATA_DIR, BatchData, Job, Batch
from ..components.simulation import Simulation

# note: this is to avoid annoying errors with asyncio.run() on jupyter notebooks
nest_asyncio.apply()


# pylint:disable=too-many-arguments, too-many-locals
async def _run_async(
    simulations: Dict[str, Simulation],
    folder_name: str = "default",
    path_dir: str = DEFAULT_DATA_DIR,
    callback_url: str = None,
    num_workers: int = None,
    verbose: bool = True,
) -> BatchData:
    """Submits a set of :class:`.Simulation` objects to server, starts running,
    monitors progress, downloads, and loads results as a :class:`.BatchData` object.
    Uses ``asyncio`` to perform these steps asynchronously.

    Parameters
    ----------
    simulations : Dict[str, :class:`.Simulation`]
        Mapping of task name to simulation.
    folder_name : str = "default"
        Name of folder to store each task on web UI.
    path_dir : str
        Base directory where data will be downloaded, by default current working directory.
    callback_url : str = None
        Http PUT url to receive simulation finish event. The body content is a json file with
        fields ``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.
    num_workers: int = None
        Number of tasks to submit at once in a batch, if None, will run all at the same time.
    verbose : bool = True
        If `True`, will print progressbars and status, otherwise, will run silently.

    Note
    ----
    This is an experimental feature and may not work on all systems or configurations.
    For more details, see ``https://realpython.com/async-io-python/``.

    Returns
    ------
    :class:`BatchData`
        Contains the :class:`.SimulationData` of each :class:`.Simulation` in :class:`Batch`.
    """

    # if number of workers not specified, just use the number of simulations
    if num_workers is None:
        num_workers = len(simulations)

    def get_path(job: Job) -> str:
        """Get the data path associated with a :class:`.Job` using :class:`.Batch` convention."""

        # pylint:disable=protected-access
        return Batch._job_data_path(task_id=job.task_id, path_dir=path_dir)

    async def worker(queue):
        """Defines a worker that runs a job."""

        while True:

            # grab a job from the queue
            job = await queue.get()

            # monitor it
            job.monitor()

            # and mark the job as completed
            queue.task_done()

    # initialize a queue
    queue = asyncio.Queue()

    # store task ids and data paths to initialize a BatchData object with for return
    task_ids = {}
    task_paths = {}

    # for each simulation provided
    for task_name, simulation in simulations.items():

        # make and start a job
        job = Job(
            simulation=simulation,
            task_name=task_name,
            callback_url=callback_url,
            folder_name=folder_name,
            verbose=verbose,
        )
        job.start()

        # store the task id and data path
        task_ids[task_name] = job.task_id
        task_paths[task_name] = get_path(job)

        # put the job on the queue
        queue.put_nowait(job)

    # create a bunch of workers to asynchonously work on the task
    tasks = []
    for _ in range(num_workers):
        task = asyncio.create_task(worker(queue))
        tasks.append(task)

    # Wait until the queue is fully processed
    await queue.join()

    # Cancel our worker tasks
    for task in tasks:
        task.cancel()

    # Wait until all worker tasks are cancelled
    await asyncio.gather(*tasks, return_exceptions=True)

    # return the batch data containing all of the job run details for loading later
    return BatchData(task_ids=task_ids, task_paths=task_paths, verbose=verbose)


# pylint:disable=too-many-arguments
def run_async(
    simulations: Dict[str, Simulation],
    folder_name: str = "default",
    path_dir: str = DEFAULT_DATA_DIR,
    callback_url: str = None,
    num_workers: int = None,
    verbose: bool = True,
) -> BatchData:
    """Submits a set of :class:`.Simulation` objects to server, starts running,
    monitors progress, downloads, and loads results as a :class:`.BatchData` object.
    Uses ``asyncio`` to perform these steps asynchronously.

    Parameters
    ----------
    simulations : Dict[str, :class:`.Simulation`]
        Mapping of task name to simulation.
    folder_name : str = "default"
        Name of folder to store each task on web UI.
    path_dir : str
        Base directory where data will be downloaded, by default current working directory.
    callback_url : str = None
        Http PUT url to receive simulation finish event. The body content is a json file with
        fields ``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.
    num_workers: int = None
        Number of tasks to submit at once in a batch, if None, will run all at the same time.
    verbose : bool = True
        If `True`, will print progressbars and status, otherwise, will run silently.

    Note
    ----
    This is an experimental feature and may not work on all systems or configurations.
    For more details, see ``https://realpython.com/async-io-python/``.

    Returns
    ------
    :class:`BatchData`
        Contains the :class:`.SimulationData` of each :class:`.Simulation` in :class:`Batch`.
    """

    return asyncio.run(
        _run_async(
            simulations=simulations,
            folder_name=folder_name,
            path_dir=path_dir,
            callback_url=callback_url,
            num_workers=num_workers,
            verbose=verbose,
        )
    )



"""Component imports for adjoint plugin. from tidy3d.plugins.adjoint.components import *"""

# import the jax version of tidy3d components
from .geometry import JaxBox  # , JaxPolySlab
from .medium import JaxMedium, JaxAnisotropicMedium, JaxCustomMedium
from .structure import JaxStructure
from .simulation import JaxSimulation
from .data.sim_data import JaxSimulationData
from .data.monitor_data import JaxModeData
from .data.dataset import JaxPermittivityDataset
from .data.data_array import JaxDataArray



"""External plugins that have tidy3d as dependency and add functionality."""

# dispersion fitter
from .dispersion.fit import DispersionFitter
from .dispersion.fit_web import StableDispersionFitter, AdvancedFitterParam

# mode solver
from .mode.mode_solver import ModeSolver, ModeSolverData

# scattering matrix
from .smatrix.smatrix import ComponentModeler, Port

# resonance finder
from .resonance.resonance import ResonanceFinder

# self-intersecting polyslab
from .polyslab.polyslab import ComplexPolySlab

# note: adjoint is imported directly from tidy3d.plugin.adjoint due to jax dependency



""" Divide a complex polyslab where self-intersecting polygon can occur during extrusion."""

from typing import List, Tuple
from math import isclose

import pydantic
from shapely.geometry import Polygon

from ...log import log
from ...components.geometry import PolySlab, GeometryGroup
from ...components.medium import MediumType
from ...components.structure import Structure
from ...components.types import Axis
from ...constants import fp_eps


# Warn for too many divided polyslabs
_WARN_MAX_NUM_DIVISION = 100


class ComplexPolySlab(PolySlab):
    """Interface for dividing a complex polyslab where self-intersecting polygon can
    occur during extrusion.

    Example
    -------
    >>> vertices = ((0, 0), (1, 0), (1, 1), (0, 1), (0, 0.9), (0, 0.11))
    >>> p = ComplexPolySlab(vertices=vertices, axis=2, slab_bounds=(0, 1), sidewall_angle=0.785)
    >>> # To obtain the divided polyslabs, there are two approaches:
    >>> # 1) a list of divided polyslabs
    >>> geo_list = p.sub_polyslabs
    >>> # 2) geometry group containing the divided polyslabs
    >>> geo_group = p.geometry_group
    >>> # Or directly obtain the structure with a user-specified medium
    >>> mat = td.Medium(permittivity=2)
    >>> structure = p.to_structure(mat)

    Note
    ----
    This version is limited to neighboring vertex-vertex crossing type of
    self-intersecting events. Extension to cover all types of self-intersecting
    events is expected in the future.

    The algorithm is as follows (for the convenience of illustration,
    let's consider the reference plane to lie at the bottom of the polyslab),

    1. Starting from the reference plane, find out the critical
    extrusion distance for the first vertices degeneracy
    event when marching towards the top of the polyslab;

    2. Construct a sub-polyslab whose base is the polygon at
    the reference plane and height to be the critical
    extrusion distance;

    3. At the critical extrusion distance, constructing a new polygon
    that keeps only one of the degenerate vertices;

    4. Set the reference plane to the position of the new polygon,
    and  repeating 1-3 to construct sub-polyslabs until reaching
    the top of the polyslab, or all vertices collapsed into a 1D curve
    or a 0D point.
    """

    @pydantic.validator("vertices", always=True)
    def no_self_intersecting_polygon_during_extrusion(cls, val, values):
        """Turn off the validation for this class."""
        return val

    @classmethod
    def from_gds(  # pylint:disable=too-many-arguments
        cls,
        gds_cell,
        axis: Axis,
        slab_bounds: Tuple[float, float],
        gds_layer: int,
        gds_dtype: int = None,
        gds_scale: pydantic.PositiveFloat = 1.0,
        dilation: float = 0.0,
        sidewall_angle: float = 0,
        **kwargs,
    ) -> List[PolySlab]:
        """Import :class:`.PolySlab` from a ``gdstk.Cell``.

        Parameters
        ----------
        gds_cell : gdstk.Cell
            ``gdstk.Cell`` containing 2D geometric data.
        axis : int
            Integer index into the polygon's slab axis. (0,1,2) -> (x,y,z).
        slab_bounds: Tuple[float, float]
            Minimum and maximum positions of the slab along ``axis``.
        gds_layer : int
            Layer index in the ``gds_cell``.
        gds_dtype : int = None
            Data-type index in the ``gds_cell``.
            If ``None``, imports all data for this layer into the returned list.
        gds_scale : float = 1.0
            Length scale used in GDS file in units of MICROMETER.
            For example, if gds file uses nanometers, set ``gds_scale=1e-3``.
            Must be positive.
        dilation : float = 0.0
            Dilation of the polygon in the base by shifting each edge along its
            normal outwards direction by a distance;
            a negative value corresponds to erosion.
        sidewall_angle : float = 0
            Angle of the sidewall.
            ``sidewall_angle=0`` (default) specifies vertical wall,
            while ``0<sidewall_angle<np.pi/2`` for the base to be larger than the top.
        reference_plane : PlanePosition = "bottom"
            The position of the GDS layer. It can be at the ``bottom``, ``middle``,
            or ``top`` of the PolySlab. E.g. if ``axis=1``, ``bottom`` refers to the
            negative side of y-axis, and ``top`` refers to the positive side of y-axis.

        Returns
        -------
        List[:class:`.PolySlab`]
            List of :class:`.PolySlab` objects sharing ``axis`` and  slab bound properties.
        """

        # TODO: change for 2.0
        # handle reference plane kwarg
        reference_plane = cls._set_reference_plane_kwarg(sidewall_angle, **kwargs)
        all_vertices = cls._load_gds_vertices(gds_cell, gds_layer, gds_dtype, gds_scale)
        polyslabs = [
            cls(
                vertices=verts,
                axis=axis,
                slab_bounds=slab_bounds,
                dilation=dilation,
                sidewall_angle=sidewall_angle,
                reference_plane=reference_plane,
            )
            for verts in all_vertices
        ]
        return [sub_poly for sub_polys in polyslabs for sub_poly in sub_polys.sub_polyslabs]

    def to_structure(self, medium: MediumType) -> Structure:
        """Construct a structure containing a user-specified medium
        and a GeometryGroup made of all the divided PolySlabs from this object.

        Parameters
        ----------
        medium : :class:`.MediumType`
            Medium for the complex polyslab.

        Returns
        -------
        :class:`.Structure`
            The structure containing all divided polyslabs made of a user-specified
            medium.
        """
        return Structure(geometry=self.geometry_group, medium=medium)

    @property
    def geometry_group(self) -> GeometryGroup:
        """Divide a complex polyslab into a list of simple polyslabs, which
        are assembled into a :class:`.GeometryGroup`.

        Returns
        -------
        :class:`.GeometryGroup`
            GeometryGroup for a list of simple polyslabs divided from the complex
            polyslab.
        """
        return GeometryGroup(geometries=self.sub_polyslabs)

    @property
    def sub_polyslabs(self) -> List[PolySlab]:
        """Divide a complex polyslab into a list of simple polyslabs.
        Only neighboring vertex-vertex crossing events are treated in this
        version.

        Returns
        -------
        List[PolySlab]
            A list of simple polyslabs.
        """
        sub_polyslab_list = []
        num_division_count = 0
        # initialize sub-polyslab parameters
        sub_polyslab_dict = self.dict(exclude={"type"}).copy()
        if isclose(self.sidewall_angle, 0):
            return [PolySlab.parse_obj(sub_polyslab_dict)]

        sub_polyslab_dict.update({"dilation": 0})  # dilation accounted in setup
        # initalize offset distance
        offset_distance = 0

        for dist_val in self._dilation_length:
            dist_now = 0.0
            vertices_now = self.reference_polygon

            # constructing sub-polyslabs until reaching the base/top
            while not isclose(dist_now, dist_val):
                # bounds for sub-polyslabs assuming no self-intersection
                slab_bounds = [
                    self._dilation_value_at_reference_to_coord(dist_now),
                    self._dilation_value_at_reference_to_coord(dist_val),
                ]
                # 1) find out any vertices touching events between the current
                # position to the base/top
                max_dist = PolySlab._neighbor_vertices_crossing_detection(
                    vertices_now, dist_val - dist_now
                )

                # vertices touching events captured, update bounds for sub-polyslab
                if max_dist is not None:
                    # max_dist doesn't have sign, so construct signed offset distance
                    offset_distance = max_dist * dist_val / abs(dist_val)
                    slab_bounds[1] = self._dilation_value_at_reference_to_coord(
                        dist_now + offset_distance
                    )

                # 2) construct sub-polyslab
                slab_bounds.sort()  # for reference_plane=top/bottom, bounds need to be ordered
                # direction of marching
                reference_plane = "bottom" if dist_val / self._tanq < 0 else "top"
                sub_polyslab_dict.update(
                    dict(
                        slab_bounds=tuple(slab_bounds),
                        vertices=vertices_now,
                        reference_plane=reference_plane,
                    )
                )
                sub_polyslab_list.append(PolySlab.parse_obj(sub_polyslab_dict))

                # Now Step 3
                if max_dist is None:
                    break
                dist_now += offset_distance
                # new polygon vertices where collapsing vertices are removed but keep one
                vertices_now = PolySlab._shift_vertices(vertices_now, offset_distance)[0]
                vertices_now = PolySlab._remove_duplicate_vertices(vertices_now)
                # all vertices collapse
                if len(vertices_now) < 3:
                    break
                # polygon collapse into 1D
                if Polygon(vertices_now).buffer(0).area < fp_eps**2:
                    break
                vertices_now = PolySlab._orient(vertices_now)
                num_division_count += 1

        if num_division_count > _WARN_MAX_NUM_DIVISION:
            log.warning(
                "Two many self-intersecting events: "
                f"The polyslab has been divided into {num_division_count} polyslabs; "
                f"any more than around {_WARN_MAX_NUM_DIVISION} may slow down the simulation."
            )

        return sub_polyslab_list

    @property
    def _dilation_length(self) -> List[float]:
        """dilation length from reference plane to the top/bottom of the polyslab."""

        # for "bottom", only needs to compute the offset length to the top
        dist = [self._extrusion_length_to_offset_distance(self.length_axis)]
        # reverse the dilation value if the reference plane is on the top
        if self.reference_plane == "top":
            dist = [-dist[0]]
        # for middle, both directions
        elif self.reference_plane == "middle":
            dist = [dist[0] / 2, -dist[0] / 2]
        return dist

    def _dilation_value_at_reference_to_coord(self, dilation: float) -> float:
        """Compute the coordinate based on the dilation value to the reference plane."""

        z_coord = -dilation / self._tanq + self.slab_bounds[0]
        if self.reference_plane == "middle":
            return z_coord + self.length_axis / 2
        if self.reference_plane == "top":
            return z_coord + self.length_axis
        # bottom case
        return z_coord



"""Defines jax-compatible MonitorData and their conversion to adjoint sources."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Union, List
import pydantic as pd
import numpy as np
import jax.numpy as jnp

from jax.tree_util import register_pytree_node_class

from .....components.source import Source, GaussianPulse
from .....components.source import ModeSource, PlaneWave, CustomFieldSource
from .....components.data.monitor_data import MonitorData
from .....components.data.monitor_data import ModeData, DiffractionData, FieldData
from .....components.data.dataset import FieldDataset
from .....components.data.data_array import ScalarFieldDataArray
from .....constants import C_0, ETA_0
from .....exceptions import AdjointError

from .data_array import JaxDataArray
from ..base import JaxObject


class JaxMonitorData(MonitorData, JaxObject, ABC):
    """A :class:`.MonitorData` that we regsiter with jax."""

    @classmethod
    def from_monitor_data(cls, mnt_data: MonitorData) -> JaxMonitorData:
        """Construct a :class:`.JaxMonitorData` instance from a :class:`.MonitorData`."""
        self_dict = mnt_data.dict(exclude={"type"}).copy()
        for field_name in cls.get_jax_field_names():
            data_array = self_dict[field_name]
            coords = {dim: data_array.coords[dim].values.tolist() for dim in data_array.coords.dims}
            jax_amps = JaxDataArray(values=data_array.values, coords=coords)
            self_dict[field_name] = jax_amps
        return cls.parse_obj(self_dict)

    @abstractmethod
    def to_adjoint_sources(self, fwidth: float) -> List[Source]:
        """Construct a list of adjoint sources from this :class:`.JaxMonitorData`."""

    @staticmethod
    def make_source_time(amp_complex: complex, freq: float, fwidth: float) -> GaussianPulse:
        """Create a :class:`.SourceTime` for the adjoint source given an amplitude and freq."""
        # fwidth = freq * FWIDTH_FACTOR
        amp = abs(amp_complex)
        phase = np.angle(1j * amp_complex)
        return GaussianPulse(freq0=freq, fwidth=fwidth, amplitude=amp, phase=phase)

    @staticmethod
    def flip_direction(direction: str) -> str:
        """Flip a direction string ('+' or '-') to its opposite value."""
        direction = str(direction)
        if direction == "+":
            return "-"
        if direction == "-":
            return "+"
        raise AdjointError(f"Given a direction of '{direction}', expected '+' or '-'.")


@register_pytree_node_class
class JaxModeData(JaxMonitorData, ModeData):
    """A :class:`.ModeData` registered with jax."""

    amps: JaxDataArray = pd.Field(
        ...,
        title="Amplitudes",
        description="Jax-compatible modal amplitude data associated with an output monitor.",
        jax_field=True,
    )

    # pylint:disable=too-many-locals
    def to_adjoint_sources(self, fwidth: float) -> List[ModeSource]:
        """Converts a :class:`.ModeData` to a list of adjoint :class:`.ModeSource`."""

        amps, sel_coords = self.amps.nonzero_val_coords
        directions = sel_coords["direction"]
        freqs = sel_coords["f"]
        mode_indices = sel_coords["mode_index"]

        adjoint_sources = []
        for amp, direction, freq, mode_index in zip(amps, directions, freqs, mode_indices):

            # TODO: figure out where this factor comes from
            k0 = 2 * np.pi * freq / C_0
            grad_const = k0 / 4 / ETA_0
            src_amp = grad_const * amp

            src_direction = self.flip_direction(str(direction))

            adj_mode_src = ModeSource(
                size=self.monitor.size,
                center=self.monitor.center,
                source_time=self.make_source_time(amp_complex=src_amp, freq=freq, fwidth=fwidth),
                mode_spec=self.monitor.mode_spec,
                direction=str(src_direction),
                mode_index=int(mode_index),
            )
            adjoint_sources.append(adj_mode_src)

        return adjoint_sources


@register_pytree_node_class
class JaxFieldData(JaxMonitorData, FieldData):
    """A :class:`.FieldData` registered with jax."""

    Ex: JaxDataArray = pd.Field(
        None,
        title="Ex",
        description="Spatial distribution of the x-component of the electric field.",
        jax_field=True,
    )
    Ey: JaxDataArray = pd.Field(
        None,
        title="Ey",
        description="Spatial distribution of the y-component of the electric field.",
        jax_field=True,
    )
    Ez: JaxDataArray = pd.Field(
        None,
        title="Ez",
        description="Spatial distribution of the z-component of the electric field.",
        jax_field=True,
    )
    Hx: JaxDataArray = pd.Field(
        None,
        title="Hx",
        description="Spatial distribution of the x-component of the magnetic field.",
        jax_field=True,
    )
    Hy: JaxDataArray = pd.Field(
        None,
        title="Hy",
        description="Spatial distribution of the y-component of the magnetic field.",
        jax_field=True,
    )
    Hz: JaxDataArray = pd.Field(
        None,
        title="Hz",
        description="Spatial distribution of the z-component of the magnetic field.",
        jax_field=True,
    )

    # pylint:disable=too-many-locals
    def to_adjoint_sources(self, fwidth: float) -> List[CustomFieldSource]:
        """Converts a :class:`.JaxFieldData` to a list of adjoint :class:`.CustomFieldSource."""

        # parse the frequency from the scalar field data
        freqs = [scalar_fld.coords["f"] for _, scalar_fld in self.field_components.items()]
        if any((len(fs) != 1 for fs in freqs)):
            raise AdjointError("FieldData must have only one frequency.")
        freqs = [fs[0] for fs in freqs]
        if len(set(freqs)) != 1:
            raise AdjointError("FieldData must all contain the same frequency.")
        freq0 = freqs[0]

        # construct the source time dependence
        src_amp = 1.0  # TODO: how to normalize?
        source_time = self.make_source_time(amp_complex=src_amp, freq=freq0, fwidth=fwidth)

        # TODO: convert self to a 'CustomCurrentSource'-like object

        # convert all of the scalar fields to ScalarFieldDataArray
        src_field_components = {}
        for name, field_component in self.field_components.items():
            values = field_component.as_ndarray
            coords = field_component.coords
            src_field_components[name] = ScalarFieldDataArray(values, coords=coords)

        # construct the CustomFieldSource and return the single instance in a list
        dataset = FieldDataset(**src_field_components)
        custom_source = CustomFieldSource(
            center=self.monitor.center,
            size=self.monitor.size,
            source_time=source_time,
            field_dataset=dataset,
        )
        return [custom_source]


@register_pytree_node_class
class JaxDiffractionData(JaxMonitorData, DiffractionData):
    """A :class:`.DiffractionData` registered with jax."""

    Er: JaxDataArray = pd.Field(
        ...,
        title="Er",
        description="Spatial distribution of r-component of the electric field.",
        jax_field=True,
    )
    Etheta: JaxDataArray = pd.Field(
        ...,
        title="Etheta",
        description="Spatial distribution of the theta-component of the electric field.",
        jax_field=True,
    )
    Ephi: JaxDataArray = pd.Field(
        ...,
        title="Ephi",
        description="Spatial distribution of phi-component of the electric field.",
        jax_field=True,
    )
    Hr: JaxDataArray = pd.Field(
        ...,
        title="Hr",
        description="Spatial distribution of r-component of the magnetic field.",
        jax_field=True,
    )
    Htheta: JaxDataArray = pd.Field(
        ...,
        title="Htheta",
        description="Spatial distribution of theta-component of the magnetic field.",
        jax_field=True,
    )
    Hphi: JaxDataArray = pd.Field(
        ...,
        title="Hphi",
        description="Spatial distribution of phi-component of the magnetic field.",
        jax_field=True,
    )

    """Note: these two properties need to be overwritten to be compatible with this subclass."""

    @property
    def amps(self) -> JaxDataArray:
        """Complex power amplitude in each order for 's' and 'p' polarizations."""

        # compute angles and normalization factors
        theta_angles = jnp.array(self.angles[0].values)
        cos_theta = np.cos(np.nan_to_num(theta_angles))
        cos_theta[cos_theta <= 0] = np.inf
        norm = 1.0 / np.sqrt(2.0 * self.eta) / np.sqrt(cos_theta)

        # stack the amplitudes in s- and p-components along a new polarization axis
        amps_phi = norm * jnp.array(self.Ephi.values)
        amps_theta = norm * jnp.array(self.Etheta.values)
        amp_values = jnp.stack((amps_phi, amps_theta), axis=3)

        # construct the coordinates and values to return JaxDataArray
        amp_coords = self.Etheta.coords.copy()
        amp_coords["polarization"] = ["s", "p"]
        return JaxDataArray(values=amp_values, coords=amp_coords)

    @property
    def power(self) -> JaxDataArray:
        """Total power in each order, summed over both polarizations."""

        # construct the power values
        power_values = jnp.abs(self.amps.values) ** 2
        power_values = jnp.sum(power_values, axis=-1)

        # construct the coordinates
        power_coords = self.amps.coords.copy()
        power_coords.pop("polarization")

        return JaxDataArray(values=power_values, coords=power_coords)

    # pylint:disable=too-many-locals
    def to_adjoint_sources(self, fwidth: float) -> List[PlaneWave]:
        """Converts a :class:`.DiffractionData` to a list of adjoint :class:`.PlaneWave`."""

        # extract the values coordinates of the non-zero amplitudes
        amp_vals, sel_coords = self.amps.nonzero_val_coords
        pols = sel_coords["polarization"]
        freqs = sel_coords["f"]
        orders_x = sel_coords["orders_x"]
        orders_y = sel_coords["orders_y"]

        # prepare some "global", "monitor-level" parameters for the source
        src_direction = self.flip_direction(self.monitor.normal_dir)
        theta_data, phi_data = self.angles

        adjoint_sources = []
        for amp, order_x, order_y, freq, pol in zip(amp_vals, orders_x, orders_y, freqs, pols):

            # select the propagation angles from the data
            angle_sel_kwargs = dict(orders_x=int(order_x), orders_y=int(order_y), f=float(freq))
            angle_theta = float(theta_data.sel(**angle_sel_kwargs))
            angle_phi = float(phi_data.sel(**angle_sel_kwargs))

            # if the angle is nan, this amplitude is set to 0 in the fwd pass, so should skip adj
            if np.isnan(angle_theta):
                continue

            # get the polarization angle from the data
            pol_angle = 0.0 if str(pol).lower() == "p" else np.pi / 2

            # TODO: understand better where this factor comes from
            k0 = 2 * np.pi * freq / C_0
            bck_eps = self.medium.eps_model(freq)
            grad_const = 0.5 * k0 / np.sqrt(bck_eps) * np.cos(angle_theta)
            src_amp = grad_const * amp

            adj_plane_wave_src = PlaneWave(
                size=self.monitor.size,
                center=self.monitor.center,
                source_time=self.make_source_time(amp_complex=src_amp, freq=freq, fwidth=fwidth),
                direction=src_direction,
                angle_theta=angle_theta,
                angle_phi=angle_phi,
                pol_angle=pol_angle,
            )
            adjoint_sources.append(adj_plane_wave_src)

        return adjoint_sources


# allowed types in JaxSimulationData.output_data
JaxMonitorDataType = Union[JaxModeData, JaxDiffractionData, JaxFieldData]

# maps regular Tidy3d MonitorData to the JaxTidy3d equivalents, used in JaxSimulationData loading
# pylint: disable=unhashable-member
JAX_MONITOR_DATA_MAP = {
    DiffractionData: JaxDiffractionData,
    ModeData: JaxModeData,
    FieldData: JaxFieldData,
}
